# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015-2016  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import pkg_resources

from . import utils
from . import processor
from . import branches
from . import instances
from . import frontend
from . import gitlab


class TestsManager:
    """Main instance of the application"""

    def __init__(self, root, config, port):
        self.config = config
        self.root = root

        self.gitlab = gitlab.GitLabAPI(self)

        self.processor = processor.Processor()
        self.frontend = frontend.create_app(self)
        self.instances = instances.InstancesManager(self, self.frontend, port)

        if not utils.is_root_valid(root):
            raise RuntimeError("Invalid root directory: %s" % root)

        self.config_file = root / "config.json"
        self.gunicorn_config_file = pkg_resources.resource_filename(
            "managetests", "gunicorn_config"
        )

        # Load the details
        with (self.root / "details.json").open() as f:
            self.details = json.load(f)

        self.init_branches()

    def init_branches(self):
        """Load all the branches"""
        self.branches = {}

        def add(name, mr):
            """Add a new branch"""
            if name in self.branches:
                return
            self.branches[name] = branches.Branch(self, name, mr)

        # Load already present branches from merge requests
        for name, mr in self.details["branches"].copy().items():
            add(name, mr)

        # Load also new forced-to-be-keeped branches
        for name in self.config["keep-branches"]:
            add(name, None)

        # Load also new branches from merge requests
        new = self.gitlab.merge_requests()
        if new is None:
            raise RuntimeError("Can't get merge requests from GitLab!")
        for request in new:
            add(request["source_branch"], request["iid"])
            self.details["branches"][request["source_branch"]] = request["iid"]

        self.save_details()

    def sync_branches_jobs(self):
        """Get the jobs needed to sync branches"""
        jobs = []

        for branch in list(self.branches.values()):
            if branch.active:
                if not branch.present():
                    # Set the branch status to "pending"
                    branch.update_status("pending")

                    # Deploy the branch and load it
                    def deploy(branch=branch):
                        branch.deploy()
                        self.instances.load_branch(branch)
                    jobs.append(deploy)
            else:
                if branch.present():
                    # Unload the branch, destroy and forget about it
                    def destroy(branch=branch):
                        self.instances.remove_branch(branch.name)
                        branch.destroy()

                        if branch.name in self.details["branches"]:
                            del self.details["branches"]
                            self.save_details()
                    jobs.append(destroy)

        return jobs

    def save_details(self):
        """Save details on disk"""
        with (self.root / "details.json").open("w") as f:
            json.dump(self.details, f)

    def run(self):
        """Run all the test instances"""
        # Load all the branches in the instances manager
        for branch in self.branches.values():
            if not branch.present():
                continue
            self.instances.load_branch(branch)

        # Queue all the sync jobs
        for job in self.sync_branches_jobs():
            self.processor.queue(job)

        # Start the various components
        self.processor.start()
        self.instances.run()

        # Stop the processor
        self.processor.stop()

    def queue_hook(self, data):
        """Queue a new hook"""
        if data["object_kind"] == "push":
            func = self._process_push_hook
        elif data["object_kind"] == "merge_request":
            func = self._process_merge_request_hook

        self.processor.queue(lambda: func(data))

    def _process_push_hook(self, data):
        """Process a push hook received from GitLab"""
        branch = data["ref"]
        if branch.startswith("refs/heads/"):
            branch = branch[len("refs/heads/"):]

        # Consider only the branches managed by managetests
        if branch not in self.branches:
            return

        # Rebuild the branch from scratch
        self.instances.remove_branch(branch)
        self.branches[branch].destroy()
        self.branches[branch].deploy()
        self.instances.load_branch(self.branches[branch])

    def _process_merge_request_hook(self, data):
        """Process a merge request hook received from GitLab"""
        obj = data["object_attributes"]
        branch = obj["source_branch"]

        # The branch was loaded but the merge request is now closed
        if branch in self.branches and obj["state"] != "opened":
            self.gitlab.post_comment(obj["id"], "L'istanza live per il branch "
                                     "**%s** è stata rimossa." % branch)

            self.instances.remove_branch(branch)
            self.branches[branch].destroy()
            del self.branches[branch]

        # The instance should be created
        elif branch not in self.branches and obj["state"] == "opened":
            self.branches[branch] = branches.Branch(self, branch, obj["iid"])
            self.branches[branch].deploy()
            self.instances.load_branch(self.branches[branch])

            self.gitlab.post_comment(
                obj["id"],
                "L'istanza live per il branch **%s** è ora "
                "[disponibile](%s/%s/)." %
                (branch, self.config["base-url"], branch)
            )
